// Turn off compiler optimizations
// Try to turn on compiler optimizations and execute again
// Measure time of operation (available to an attacker)
// Measure counts of operation (usually not available to an atatcker, but gives us insight)
// Try profiler to obtain count of operations

// What is "secret" information stored at server? What is data that attacker can choose?

// Plot.ly - Traces->Range/bins 1

#define _CRT_SECURE_NO_WARNINGS

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <memory.h>
#include "bignum.h"

typedef unsigned long ULONG;
const int ULONG_LENGTH = sizeof(ULONG);

typedef enum {
	NAIVE = 1,
	SQUAREANDMULTIPLY = 2,
	SQUAREANDMULTIPLYMASKED = 3,
	SLIDINGWINDOWEXPMOD = 4
} ExpType;
 
// Note: we will use long int for easy manipulation with variables. However, length unsigned int (32 bits) is not enough to provide secure RSA!

/**
 x^c mod n => ((x*x) mod n * x) mod n) * x mod n) ...  => c-1 modulo exponentiation
*/
ULONG naiveExponentiation(ULONG message, ULONG exponent, ULONG modulus) {
	ULONG result = message;
	for (ULONG i = 1; i < exponent; i++) {
		// result = (result * message) % modulus; // this may cause type overflow
		result *= message;
		result %= modulus;
	}

	return result;
}


ULONG squareAndMultiply(ULONG message, ULONG exponent, ULONG modulus) {
	// Obtain effective length of exponent in bits
	int sizeExponent = ULONG_LENGTH;
	ULONG mask = 1;
	ULONG bit = 0;
	for (int i = 0; i < ULONG_LENGTH * 8; i++) {
		bit = exponent & mask;
		if (bit != 0) { sizeExponent = i + 1; }
		mask <<= 1;
	}

	// Compute square and multiply algorithm
	ULONG result = 1;
	for (int i = sizeExponent - 1; i >= 0; i--) {
		//result = (result * result) % modulus; // this may cause type overflow
		result *= result;
		result %= modulus;
		if ((exponent & (1 << i)) != 0) { // given bit is not 0
			// result = (result * message) % modulus; // this may cause type overflow
			result *= message;
			result %= modulus;
		}
	}

	return result;
}

void naiveExponentiationMPI(const mpi* message, const mpi* exponent, const mpi* modulus, mpi* result) {
	mpi_copy(result, message);

	mpi counter; mpi_init(&counter);
	mpi_copy(&counter, exponent);

	while (mpi_cmp_int(&counter, 1) > 0) {
		mpi_mul_mpi(result, result, message);
		mpi_mod_mpi(result, result, modulus);
		mpi_sub_int(&counter, &counter, 1);
	}
}

void squareAndMultiplyMPI(const mpi* message, const mpi* exponent, const mpi* modulus, mpi* result) {
	// Obtain length of exponent in bits (shoudl be ideally pre-computed)
	int sizeExponent = 0;
	int bit = 0;
	int maxBitLength = mpi_size(exponent) * 8;
	for (int i = 0; i < maxBitLength; i++) {
		bit = mpi_get_bit(exponent, i);
		if (bit != 0) { sizeExponent = i + 1; }
	}

	// Compute square and multiply algorithm
	mpi_lset(result, 1);
	for (int i = sizeExponent - 1; i >= 0; i--) {
		mpi_mul_mpi(result, result, result); // result *= result;
		mpi_mod_mpi(result, result, modulus); // result %= modulus;
		if (mpi_get_bit(exponent, i) != 0) { // given bit is not 0
    		// result = (result * message) % modulus;
			mpi_mul_mpi(result, result, message);
			mpi_mod_mpi(result, result, modulus); 
		}
	}
}

long measureExponentiation(ULONG message, ULONG exponent, ULONG modulus, ExpType exponentiationType) {
	ULONG result = 0;
	clock_t begin = clock();
	switch (exponentiationType) {
		case NAIVE: {
			result = naiveExponentiation(message, exponent, modulus);
			break;
		}
		case SQUAREANDMULTIPLY: {
			result = squareAndMultiply(message, exponent, modulus);
			break;
		}
	}
	clock_t end = clock();
	long elapsed = end - begin;

	printf("\n######################################\n");
	if (exponentiationType == NAIVE) { printf("Method: NAIVE\n"); }
	if (exponentiationType == SQUAREANDMULTIPLY) { printf("Method: SQUAREANDMULTIPLY\n"); }
	if (exponentiationType == SQUAREANDMULTIPLYMASKED) { printf("Method: SQUAREANDMULTIPLYMASKED\n"); }
	printf("%d^%d mod %d == %d\n", message, exponent, modulus, result);
	printf("Elapsed CPU ticks: %d\n", elapsed);

	return elapsed;
}

long measureExponentiationMPI(const mpi* message, const mpi* exponent, const mpi* modulus, ExpType exponentiationType) {
	mpi result; 
	mpi_init(&result);
	clock_t begin = clock();
	switch (exponentiationType) {
	case NAIVE: {
		naiveExponentiationMPI(message, exponent, modulus, &result);
		break;
	}
	case SQUAREANDMULTIPLY: {
		squareAndMultiplyMPI(message, exponent, modulus, &result);
		break;
	}
	case SLIDINGWINDOWEXPMOD: {
		// Native MPI function
		mpi_exp_mod(&result, message, exponent, modulus, NULL);
		break;
	}
/*
	case SQUAREANDMULTIPLYMASKED: {
		result = squareAndMultiplyMaskedMPI(message, exponent, modulus);
		break;
	}
*/
	}
	clock_t end = clock();
	long elapsed = end - begin;

	printf("\n######################################\n");
	if (exponentiationType == NAIVE) { printf("Method: NAIVE MPI\n"); }
	if (exponentiationType == SQUAREANDMULTIPLY) { printf("Method: SQUAREANDMULTIPLY MPI\n"); }
	if (exponentiationType == SQUAREANDMULTIPLYMASKED) { printf("Method: SQUAREANDMULTIPLYMASKED MPI\n"); }
	if (exponentiationType == SLIDINGWINDOWEXPMOD) { printf("Method: SLIDINGWINDOWEXPMOD MPI\n"); }

	// print results
	mpi_write_file("Message\t= ", message, 16, stdout);
	mpi_write_file("Exponent= ", exponent, 16, stdout);
	mpi_write_file("Modulus\t= ", modulus, 16, stdout);
	mpi_write_file("Result\t= ", &result, 16, stdout);

	printf("Elapsed CPU ticks: %d\n", elapsed);

	return elapsed;
}



void measureExponentiationRepeat(ULONG message, ULONG exponent, ULONG modulus, ExpType exponentiationType, int numRepeats, const char* fileName) {
	FILE* outFile = fopen(fileName, "w");
	long elapsed = 0;
	for (int i = 0; i < numRepeats; i++) {
		elapsed = measureExponentiation(message, exponent, modulus, exponentiationType);
		fprintf(outFile, "%d\n", elapsed);
	}
	fclose(outFile);
}

void measureExponentiationRepeatMPI(mpi* message, mpi* exponent, mpi* modulus, ExpType exponentiationType, int numRepeats, const char* fileName) {
	FILE* outFile = fopen(fileName, "w");
	long elapsed = 0;
	for (int i = 0; i < numRepeats; i++) {
		elapsed = measureExponentiationMPI(message, exponent, modulus, exponentiationType);
		fprintf(outFile, "%d\n", elapsed);
	}
	fclose(outFile);
}

/**
	Toy random number generation function. Not suitable for rea crypto use due to weak RNG generator!!!
*/
int generateRNG(void *, unsigned char * buffer, size_t numBytes) {
	for (size_t i = 0; i < numBytes; i++) {
		buffer[i] = rand() % 256;
	}
	return 0;
}


int main()
{
	printf("sizeof(ULONG) = %d\n", sizeof(ULONG));
	printf("CLOCKS_PER_SEC = %d\n", CLOCKS_PER_SEC);
	printf("######################################\n");

	// Task 1: Try to find attack against naive multiplication based on timings
	measureExponentiation(2, 3, 1003, NAIVE);
	measureExponentiation(11, 30, 101, NAIVE);

	return 0;

	// Task 2: Try to repeat with square and multiply algorithm - does it work?
	measureExponentiation(2, 3, 1003, SQUAREANDMULTIPLY);
	measureExponentiation(11, 30, 101, SQUAREANDMULTIPLY);

	// Task 3: How big number we can multiply with integer type properly?
	measureExponentiation(65535, 65535, 100000003, NAIVE);  // extremely slow
	measureExponentiation(65535, 65535, 100000003L, SQUAREANDMULTIPLY); // very fast, but timing problem

	// Obtain statistics from 100 repeats for same repeated calls - how does it compare?
	measureExponentiationRepeat(10, 65535, 10000003L, NAIVE, 100, "measurement_naive1.txt");
	measureExponentiationRepeat(10, 65535, 10000003L, NAIVE, 100, "measurement_naive2.txt");

	// Use calc to create exponent with hig/low hamming weight
	//measureExponentiationRepeat(12345678L, 268435455L, 10000003L, NAIVE, 100, "measurement_naive_highHW.txt"); // 268435455L == 1111...11 (28 bits total)
	//measureExponentiationRepeat(12345678L, 134217728L, 10000003L, NAIVE, 100, "measurement_naive_lowHW.txt"); // 134217728L == 100....00 (28 bits total)
	//measureExponentiationRepeat(12345678L, 268435455L, 10000003L, SQUAREANDMULTIPLY, 100, "measurement_sqandm_highHW.txt"); // 268435455L == 1111...11 (28 bits total)
	//measureExponentiationRepeat(12345678L, 134217728L, 10000003L, SQUAREANDMULTIPLY, 100, "measurement_sqandm_lowHW.txt"); // 134217728L == 100....00 (28 bits total)


	//
	// Task 4: use big number data type - MPI
	// 
	mpi message; mpi_init(&message);
	mpi exponent; mpi_init(&exponent);
	mpi modulus; mpi_init(&modulus);

	// Small number for testing
	measureExponentiation(11L, 2L, 101L, NAIVE);
	mpi_lset(&message, 11); mpi_lset(&exponent, 2); mpi_lset(&modulus, 101);
	measureExponentiationMPI(&message, &exponent, &modulus, NAIVE);

	// Relatively large numbers - but still too fast to measure with clock() on current CPUs
	mpi_lset(&message, 12345678L);
	mpi_lset(&exponent, 268435455L);
	mpi_lset(&modulus, 1000000003L);
	//measureExponentiationMPI(&message, &exponent, &modulus, NAIVE); // too long
	measureExponentiationMPI(&message, &exponent, &modulus, SQUAREANDMULTIPLY);
	measureExponentiationMPI(&message, &exponent, &modulus, SLIDINGWINDOWEXPMOD); // native MPI implementation via sliding window method

	// Task 5: Measurements with cryptographically large numbers (2048b)
	const int NUMBER_SIZE = 256;
	// Init with pseudorandom values (prng will always start with same value)
	mpi_fill_random(&message, NUMBER_SIZE, generateRNG, NULL);
	mpi_fill_random(&exponent, NUMBER_SIZE, generateRNG, NULL);
	mpi_fill_random(&modulus, NUMBER_SIZE, generateRNG, NULL);
	// Fix MSb and LSb of modulus to 1
	modulus.p[0] |= 1; mpi_set_bit(&modulus, 1, 1);
	 
	measureExponentiationMPI(&message, &exponent, &modulus, SQUAREANDMULTIPLY);
	measureExponentiationMPI(&message, &exponent, &modulus, SLIDINGWINDOWEXPMOD);
	// Why our SQUAREANDMULTIPLY is slower than PolarSSL's SLIDINGWINDOWEXPMOD?

	//
	// Task 6: Dependency of timing on hamming weight of exponent
	//
	mpi lowHWExponent; mpi_init(&lowHWExponent);
	mpi highHWExponent; mpi_init(&highHWExponent);
	mpi_copy(&lowHWExponent, &exponent);
	mpi_copy(&highHWExponent, &exponent);
	for (int i = 2; i < (NUMBER_SIZE * 7) - 1; i++) { // significant part of exponent's bits is zero 
		mpi_set_bit(&lowHWExponent, i, 0);
	}
	for (int i = 2; i < (NUMBER_SIZE * 8) - 1; i++) { // all bits are 1
		mpi_set_bit(&highHWExponent, i, 1);
	}
	// Evaluate variations in distribution of time measurements between low and high hamming weight exponents
	measureExponentiationMPI(&message, &lowHWExponent, &modulus, SQUAREANDMULTIPLY);
	measureExponentiationMPI(&message, &highHWExponent, &modulus, SQUAREANDMULTIPLY);
	measureExponentiationRepeatMPI(&message, &lowHWExponent, &modulus, SQUAREANDMULTIPLY, 100, "measurement_sqandm_lowHW.txt");
	measureExponentiationRepeatMPI(&message, &highHWExponent, &modulus, SQUAREANDMULTIPLY, 100, "measurement_sqandm_highHW.txt");


	// Task 7: analysis of sliding windows 
	//measureExponentiationMPI(&message, &lowHWExponent, &modulus, SLIDINGWINDOWEXPMOD);
	//measureExponentiationMPI(&message, &highHWExponent, &modulus, SLIDINGWINDOWEXPMOD);
	//measureExponentiationRepeatMPI(&message, &lowHWExponent, &modulus, SLIDINGWINDOWEXPMOD, 100, "measurement_sqandmNative_lowHW.txt");
	//measureExponentiationRepeatMPI(&message, &highHWExponent, &modulus, SLIDINGWINDOWEXPMOD, 100, "measurement_sqandmNative_highHW.txt");
	
	return 0;
}

